{***********************************************************************
Source file: AboutBoxUnit.pas
Contains the implementation of TAboutBox, an about box that takes its
information from the executable.

Author: Niels Dekker
E-mail: dekkerware at xs4all.nl

Tested on Delphi 5, 6, 7
************************************************************************
{* Updates:
2002.08.30
  Created based on the unit QuirtAboutBox (update 20020503) that
  I made at the RT department of the Antoni van Leeuwenhoek hospital,
  with help from Marcel van Herk, Bob Brand, and Lennert Ploeger.
2002.11.11
  In ExploreFile(), Windows Explorer failed if the passed filename
  had any comma's. This was fixed by using double-quotes.
  In .DFM file: Anchors set, so that the about box will look much
  nicer when it is resized (e.g., by Maxximum Software's WinSize)
2002.11.13
  "Explore" renamed by "Find".  Some useless code removed.
2002.11.18
  In .DFM file: About box made slightly smaller
2003.07.15
  Gone freeware.
2003.07.17
  New e-mail address: dekkerware at xs4all.nl
2003.08.15
  Removed DesignSize from DFM-file, using a tool made by
  Gerrit Beuze (ModelMaker Tools): D7 DesignSize cleaner
2003.09.27
  Update history restyled.
*}
unit AboutBoxUnit;


{*
  USAGE
    Just add this file to your Delphi project (Project -> Add to Project) and
    add a button or a menu-item to your form that does:
      AboutBox.ShowModal
*}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls, Spin, Menus;

type
  TAboutBox = class(TForm)
    ButtonClose: TButton;
    ImageApplication: TImage;
    LabelApplicationName: TLabel;
    RichEditApplication: TRichEdit;
    LabelApplicationDescription: TLabel;
    PopupMenu1: TPopupMenu;
    MenuItemCopy: TMenuItem;
    MenuItemSelectAll: TMenuItem;
    N1: TMenuItem;
    MenuItemFindThisApplication: TMenuItem;
    MenuItemProperties: TMenuItem;
    procedure FormShow(Sender: TObject);
    procedure RichEditApplicationContextPopup(Sender: TObject;
      MousePos: TPoint; var Handled: Boolean);
    procedure MenuItemSelectAllClick(Sender: TObject);
    procedure MenuItemCopyClick(Sender: TObject);
    procedure MenuItemFindThisApplicationClick(Sender: TObject);
    procedure MenuItemPropertiesClick(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
{
    procedure RichEditApplicationURLClick(Sender: TObject;
      const URLText: String; Button: TMouseButton);
}
  private
    FAdditionalVersionInfo: string;
    procedure ShowApplicationVersionInfo;
  public
    property AdditionalVersionInfo: string write FAdditionalVersionInfo;
  end;

var
  AboutBox: TAboutBox;


implementation

{$R *.DFM}

uses
  ShellApi;


// Returns the double-quoted version of a string.
function DoubleQuotedStr(const s: string): string;
begin
  if Pos('"', s) = 0 then
    // The string will be enclosed by double-quotes:
    result := '"' + s + '"'
  else
    // The string already has some double-quote(s)!
    result := s
end;


// Displays the Windows properties dialog of the specified file.
// Implemented according to http://ddelphi.hypermart.net/delphi/winapi.html#prop
function DisplayPropertiesDialog(const Filename: string): boolean;
var
  Info: TShellExecuteInfo;
begin
  ZeroMemory(@Info, SizeOf(Info) );
  Info.cbSize := SizeOf(Info);
  Info.lpFile := PChar(Filename);
  Info.lpVerb := 'properties';
  Info.fMask := SEE_MASK_INVOKEIDLIST;
  result := ShellExecuteEx(@Info)
end;


function GetVersionInfoValue(
  const strFile: string;
  const strKey: string):
  string;
var
  dwHandle, dwSize: DWORD;
  pszFilename: PChar;
  aVersionInfo: array of byte;
  TranslationTable: Pointer;
  pszValue: PChar;
  uLen: UINT;
begin
  pszFilename := PChar(strFile);
  dwSize := GetFileVersionInfoSize(pszFilename, dwHandle);

  SetLength(aVersionInfo, dwSize);

  result := '';
  if GetFileVersionInfo(
    pszFilename,
    dwHandle,
    dwSize,
    aVersionInfo) then
    if VerQueryValue(
      aVersionInfo,
      '\VarFileInfo\Translation',
      TranslationTable, uLen) then

      if VerQueryValue(
        aVersionInfo,
        PChar(
          '\StringFileInfo\' +
          Format('%.4x', [LoWord(TranslationTable^)]) +
          Format('%.4x', [HiWord(DWORD(TranslationTable^))]) +
          '\' + strKey),
        Pointer(pszValue),
        uLen) then
          result := pszValue
end;


function SkipPadding(const dwIndex: DWORD): DWORD;
begin
  result := dwIndex;

  if (result mod 4) <> 0 then
    Inc(result, 4 - (dwIndex mod 4) )
end;


// Converts a pointer to a string.
function PointerToString(
  const p: Pointer;
  const iCharSize: integer): string;
begin
  case iCharSize of
    SizeOf(WideChar):
      result := WideCharToString( PWideChar(p) );
    SizeOf(AnsiChar):
      result := string( PAnsiChar(p) )
  else
    result := string( PChar(p) )
  end;
end;


function FindStringTable(
  out dwStringTable: DWORD;
  const aVersionInfo: array of byte;
  const iCharSize: integer):
  boolean;
const
  KEY = 'StringFileInfo';
var
  iIndex: integer;
begin
  for iIndex := iCharSize * Length(KEY) to Length(aVersionInfo) - 1 do
    if ((iCharSize = SizeOf(WideChar)) and (PWideChar(@aVersionInfo[iIndex])^ = WideChar(#0))) or
       ((iCharSize = SizeOf(AnsiChar)) and (PAnsiChar(@aVersionInfo[iIndex])^ = AnsiChar(#0))) then
    begin
      if PointerToString(
        @aVersionInfo[iIndex - Length(KEY) * iCharSize],
        iCharSize) = KEY then
      begin
        dwStringTable := SkipPadding(iIndex);
        result := true;
        exit
      end
    end;
  result := false
end;


procedure AppendVersionInfoStrings(
  const rStrings: TStrings;
  const strFile: string);
var
  dwHandle, dwSize: DWORD;
  pszFilename: PChar;
  aVersionInfo: array of byte;
  dwStringTable, dwString, dwKey, dwValue: DWORD;
  wLength: WORD;
  strKey, strValue: string;
  iCharSize: integer;
begin
  pszFilename := PChar(strFile);
  dwSize := GetFileVersionInfoSize(pszFilename, dwHandle);
  SetLength(aVersionInfo, dwSize);

  if GetFileVersionInfo(
    pszFilename,
    dwHandle,
    dwSize,
    aVersionInfo) then
  begin
    // Usually the version information that was retrieved from GetFileVersionInfo
    // is stored in wide (Unicode) character strings...
    iCharSize := SizeOf(WideChar);
    if not FindStringTable(dwStringTable, aVersionInfo, iCharSize) then
    begin
      // ...But if you have Win95, the string tables inside the retrieved
      // version information may contain ordinary single-byte (Ansi) characters!
      iCharSize := SizeOf(AnsiChar);
      if not FindStringTable(dwStringTable, aVersionInfo, iCharSize) then
        exit
    end;

    dwString := dwStringTable + 8 + DWORD(8 * iCharSize);
    wLength := PWORD(@aVersionInfo[dwStringTable])^;

    while (dwString + 6) < (dwStringTable + wLength) do
    begin
      dwKey := dwString + 2 + DWORD(2 * iCharSize);
      strKey := PointerToString(@aVersionInfo[dwKey], iCharSize);
      dwValue := SkipPadding(dwKey + DWORD((Length(strKey) + 1) * iCharSize));
      strValue := PointerToString(@aVersionInfo[dwValue], iCharSize);
      rStrings.Append(strKey + '=' + strValue);
      dwString := SkipPadding(dwValue + DWORD((Length(strValue) + 1) * iCharSize));
    end
  end
end;


// Returns the modified date and time of the specified file as a human-friendly
// readable string.  Or an empty string if something goes wrong.
// Uses the Windows API FindFirstFile() instead of the VCL function FileAge(),
// because the latter returns an inaccurate DOS date-and-time value.
function GetModifiedDateAndTimeOfFile(const strFile: string): string;
var
  FindData: TWin32FindData;
  SystemTime: TSystemTime;
  hFindFile: THandle;
begin
  try
    hFindFile := Windows.FindFirstFile(PChar(strFile), FindData);
    if hFindFile = INVALID_HANDLE_VALUE then
      // The specified file does not exist!
      result := ''
    else
    begin
      Windows.FindClose(hFindFile);
      if FileTimeToLocalFileTime(FindData.ftLastWriteTime, FindData.ftLastWriteTime) then
        if FileTimeToSystemTime(FindData.ftLastWriteTime, SystemTime) then
            result := DateTimeToStr(SystemTimeToDateTime(SystemTime))
        else
          // FileTimeToLocalFileTime failed.
          result := ''
      else
        // FileTimeToSystemTime failed. Maybe ftLastWriteTime >= 0x8000000000000000
        result := ''
    end
  except
    // SystemTimeToDateTime might have raised an EConvertError!  :-(
    result := ''
  end
end;


procedure StartExplorerAndSelectFile(strFile: string);
begin
  strFile := Trim(strFile);

  // If you want a file to be selected inside the Windows Explorer and the
  // filename contains a comma, the filename needs to have double-quotes.
  // For both Windows NT4 SP6 and Windows 98 SE the following command will fail:
  //   explorer.exe /select,C:\Filename,ThatContainsAComma.txt
  // But the following should work correctly:
  //   explorer.exe /select,"C:\Filename,ThatContainsAComma.txt"
  if Pos(strFile, ',') <> 0 then
    strFile := DoubleQuotedStr(strFile);

  ShellExecute(
    0,
    'open',
    PChar('explorer.exe'),
    PChar('/n,/select,' + strFile),
    nil,
    SW_SHOWNORMAL)
end;


// This function assumes that the strings are of the form Name=Value. It locates
// the (first) string with the specified Name, retrieves its Value and deletes
// this string from the list of strings.  Returns true iff a non-empty Value is
// successfully retrieved.
function GetValueAndDelete(
  out strValue: string;
  const rStrings: TStrings;
  const strName: string):
  boolean;
var
  iIndex: integer;
begin
  iIndex := rStrings.IndexOfName(strName);
  if iIndex >= 0 then
  begin
    strValue := rStrings.Values[strName];
    result := strValue <> '';
    rStrings.Delete(iIndex)
  end
  else
    result := false
end;


// This procedure assumes that the source parameter contains strings of the
// form Name=Value.  It adds the strings that have a non-empty Value to the
// destination parameter.  The added strings will be of the form Name: Value
procedure AddStringsWithValue(const rDestination, rSource: TStrings);
var
  iIndex: integer;
  strName, strValue: string;
begin
  for iIndex := 0 to rSource.Count - 1 do
  begin
    strName := rSource.Names[iIndex];
    strValue := rSource.Values[strName];
    if strValue <> '' then
      rDestination.Append(strName + ': ' + strValue)
  end
end;


////////////////////////////////////////////////////////////////////////
// Private Methods


procedure TAboutBox.ShowApplicationVersionInfo;
var
  VersionInfo: TStringList;
  strValue: string;
  strApplicationFile: string;
  strApplicationName: string;
begin
  strApplicationFile := Application.ExeName;
  VersionInfo := TStringList.Create;
  AppendVersionInfoStrings(VersionInfo, strApplicationFile);

  if GetValueAndDelete(strValue, VersionInfo, 'ProductName') then
     strApplicationName := strValue
  else
  begin
    strApplicationName := Application.Title;

    if strApplicationName = '' then
      strApplicationName := ChangeFileExt(Application.ExeName, '')
  end;
  Caption := 'About ' + strApplicationName;
  LabelApplicationName.Caption := strApplicationName;

  if GetValueAndDelete(strValue, VersionInfo, 'FileDescription') then
    LabelApplicationDescription.Caption := strValue;

  RichEditApplication.Lines.Clear;
  RichEditApplication.Lines.Append(
    'Application file ' + DoubleQuotedStr(strApplicationFile) );

  if GetValueAndDelete(strValue, VersionInfo, 'FileVersion') then
    RichEditApplication.Lines.Append('File version: ' + strValue);

  if GetValueAndDelete(strValue, VersionInfo, 'ProductVersion') then
    RichEditApplication.Lines.Append('Product version: ' + strValue);

  RichEditApplication.Lines.Append(
    'Updated at ' +
    GetModifiedDateAndTimeOfFile(Application.ExeName));

  AddStringsWithValue(RichEditApplication.Lines, VersionInfo);
  VersionInfo.Free;

  if FAdditionalVersionInfo <> '' then
    RichEditApplication.Lines.Append(FAdditionalVersionInfo)
end;


////////////////////////////////////////////////////////////////////////
// Event Handlers


procedure TAboutBox.FormShow(Sender: TObject);
begin
  ImageApplication.Picture.Icon := Application.Icon;
  ShowApplicationVersionInfo();
end;


procedure TAboutBox.RichEditApplicationContextPopup(Sender: TObject;
  MousePos: TPoint; var Handled: Boolean);
begin
  MenuItemFindThisApplication.Caption :=
    'Find ' + DoubleQuotedStr( ExtractFileName(Application.ExeName) )
end;

procedure TAboutBox.MenuItemSelectAllClick(Sender: TObject);
begin
  RichEditApplication.SelectAll()
end;

procedure TAboutBox.MenuItemCopyClick(Sender: TObject);
begin
  RichEditApplication.CopyToClipboard()
end;

procedure TAboutBox.MenuItemFindThisApplicationClick(Sender: TObject);
begin
  StartExplorerAndSelectFile(Application.ExeName)
end;

procedure TAboutBox.MenuItemPropertiesClick(Sender: TObject);
begin
  DisplayPropertiesDialog(Application.ExeName)
end;


procedure TAboutBox.PopupMenu1Popup(Sender: TObject);
var
  iSelLength: integer;
begin
  iSelLength := RichEditApplication.SelLength;
  MenuItemSelectAll.Enabled := iSelLength < Length(RichEditApplication.Text);
  MenuItemCopy.Enabled := iSelLength > 0
end;


end.

