{***********************************************************************
Source file: Unit1.pas
Contains the implementation of the main form of AutoDrop.exe

Tested on Delphi 6, 7

Author: Niels Dekker
E-mail: dekkerware at xs4all.nl
Please drop by the website http://www.xs4all.nl/~nd/dekkerware/autodrop
************************************************************************
Updates:
2003-08-30
  Created
2003-09-03
  Released on http://www.xs4all.nl/~nd/dekkerware/autodrop
2003-09-07
  Added option "Drop Only On Top-level Windows".
  Fixed declaration of message handler WMDropFiles.
  ShortCut for menu item "Windows Explorer" set to Ctrl+E.
  Zero-sized windows are no longer a target when the
  option "Drop Only On Visible Windows" is checked.
  An instance of this application will no longer have its
  own windows as a target.
2003-09-08
  "Drop it now!" is no longer the default button (unless
  the "File to drop" edit box has the focus).  So now when
  ShellListView has the focus, entering a selected folder
  by pressing Enter is supported.
  SetPath() call replaced by simply setting Path property.
  Fixed test on whether a window is "its own window".
2003-09-09
  File edit box synchronized with any change of folder.
2003-09-14
  Fixed option "Drop Only On Visible Windows".
  Synchronization of "File to drop" edit box to Shell controls
  switched off when the edit box is changed by the user.
  View menu item "Hidden Files and Folders" added.
2003-09-25
  Version 1.0.1.0:
  Added "Window handle" combo box.
  Added "Refresh" menu item to View menu.
  Removed "Drop on child windows too" from Options menu.
  Fixed support for one commandline parameter (the file).
  Fix: OnChange handlers now nil until first idle time.
  Status bar height set to 21 and removed its outer bevel.
  Comments of EnumProc functions fixed.
  Disallowed dropping when multiple windows match the text
  Disabled the main form until it is ready.
  The path of INI file is now mentioned in the About Box.
2003-09-27
  Fixed HexToHWND-bug leading to incorrect calculation of Window handle.
  Added error message box when user specified window handle is not in list.
  Useless OnKeyDown handlers removed.  (Pressing the ENTER key when the
  "File to drop" edit box has the focus will still trigger a drop,
  because by then the "Drop it now" button will be the default button.)
  Update history restyled.
2003-10-12
  Added error message box when dropping a file by clicking
  "Drop it now!" (button or menu item) fails.
  Fixed bug: when changing the selection from one file to another while
  "Automatically drop selected file" was checked, the newly selected
  file was dropped twice instead of once.
  Comment fixed in implementation of DropFileOnWindow function.
  Unused callback function EnumProcToDropFileOnWindow removed.
2003-10-22
  In DFM file: PanelLeft.Constraints set from 0 to 1,
  SplitterVertical.MinSize set from 30 to 1, and
  panel sizes are added to FormStorage1 so that the position of the
  splitters will be restored when the user restarts AutoDrop.
  "Read Only" item added to View menu.
  Improved synchronization of list view to "File to drop" edit box.
2004-05-19
  Minor status bar fix.
  Memory leak in DropFileOnWindow fixed by calling GlobalFree.
  In DFM File: Window text combo box sorted alphabetically.
}
unit Unit1;

interface

// Please remove the following directive if you don't have RxLib installed.
{$DEFINE USE_RX_LIBRARY}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ExtCtrls, ComCtrls, ShellCtrls, StdCtrls,
{$IFDEF USE_RX_LIBRARY}
  Placemnt,  // For RxLib component TFormStorage.
{$ENDIF}
  AppEvnts, Buttons;

type
  TForm1 = class(TForm)
    ApplicationEvents1: TApplicationEvents;
    ButtonDrop: TButton;
    CheckBoxAutomatically: TCheckBox;
    ComboBoxWindow: TComboBox;
    EditFile: TEdit;
    FormStorage1: TFormStorage;
    Label1: TLabel;
    Label2: TLabel;
    MainMenu1: TMainMenu;
    MenuFile: TMenuItem;
    MenuView: TMenuItem;
    MenuOptions: TMenuItem;
    MenuHelp: TMenuItem;
    MenuItemSmallIcons: TMenuItem;
    MenuItemList: TMenuItem;
    MenuItemDetails: TMenuItem;
    MenuItemLargeIcons: TMenuItem;
    Memo1: TMemo;
    MenuItemDrop: TMenuItem;
    MenuItemAbout: TMenuItem;
    MenuItemFolders: TMenuItem;
    MenuItemCaseSensitiveWindowText: TMenuItem;
    MenuItemMatchWholeWindowTextOnly: TMenuItem;
    MenuItemExit: TMenuItem;
    MenuItemDropOnVisibleWindowsOnly: TMenuItem;
    MenuItemExplore: TMenuItem;
    MenuItemStatusBar: TMenuItem;
    MenuItemSeparator1: TMenuItem;
    MenuItemSeparator2: TMenuItem;
    MenuItemSeparator3: TMenuItem;
    PanelBottom: TPanel;
    PanelClient: TPanel;
    PanelLeft: TPanel;
    PanelTop: TPanel;
    ShellComboBox1: TShellComboBox;
    ShellTreeView1: TShellTreeView;
    ShellListView1: TShellListView;
    SplitterVertical: TSplitter;
    SplitterHorizontal: TSplitter;
    MenuItemWebsite: TMenuItem;
    MenuItemOpen: TMenuItem;
    MenuItemSeparator4: TMenuItem;
    MenuItemBrowse: TMenuItem;
    OpenDialog1: TOpenDialog;
    MenuItemDropOnlyOnToplevelWindows: TMenuItem;
    MenuItemHiddenFilesAndFolders: TMenuItem;
    ComboBoxHandle: TComboBox;
    Label3: TLabel;
    N1: TMenuItem;
    MenuItemRefresh: TMenuItem;
    MenuItemReadOnly: TMenuItem;
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure ButtonDropClick(Sender: TObject);
    procedure ComboBoxWindowDropDown(Sender: TObject);
    procedure EditFileChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure MenuViewClick(Sender: TObject);
    procedure MenuItemAboutClick(Sender: TObject);
    procedure MenuItemDetailsClick(Sender: TObject);
    procedure MenuItemDropClick(Sender: TObject);
    procedure MenuItemDropOnVisibleWindowsOnlyClick(Sender: TObject);
    procedure MenuItemExitClick(Sender: TObject);
    procedure MenuItemExploreClick(Sender: TObject);
    procedure MenuItemFoldersClick(Sender: TObject);
    procedure MenuItemLargeIconsClick(Sender: TObject);
    procedure MenuItemListClick(Sender: TObject);
    procedure MenuItemSmallIconsClick(Sender: TObject);
    procedure MenuItemStatusBarClick(Sender: TObject);
    procedure PanelLeftResize(Sender: TObject);
    procedure ShellListView1Change(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure WMDropFiles(var Msg: TWMDropFiles); message WM_DROPFILES;
    procedure MenuItemWebsiteClick(Sender: TObject);
    procedure MenuItemOpenClick(Sender: TObject);
    procedure MenuItemBrowseClick(Sender: TObject);
    procedure ComboBoxWindowChange(Sender: TObject);
    procedure MenuItemMatchWholeWindowTextOnlyClick(Sender: TObject);
    procedure MenuItemCaseSensitiveWindowTextClick(Sender: TObject);
    procedure MenuItemDropOnlyOnToplevelWindowsClick(Sender: TObject);
    procedure EditFileEnter(Sender: TObject);
    procedure EditFileExit(Sender: TObject);
    procedure ShellTreeView1Change(Sender: TObject; Node: TTreeNode);
    procedure MenuItemHiddenFilesAndFoldersClick(Sender: TObject);
    procedure ComboBoxHandleChange(Sender: TObject);
    procedure MenuItemRefreshClick(Sender: TObject);
    procedure ComboBoxHandleExit(Sender: TObject);
    procedure MenuItemReadOnlyClick(Sender: TObject);
  private
    FOverwriteStatusText: Boolean;
    FOldWidthOfLeftPanel: Integer;
    FDefaultHeightOfStatusBar: Integer;
    function SyncShellListViewWithFile(const strFile: string): string;
    procedure Drop(const strFile: string; const bShowErrorMessageBox: Boolean);
    procedure AddListOfMatchingWindowHandlesToComboBox;
    procedure AddListOfWindowsToComboBox;
  end;

var
  Form1: TForm1;

implementation

uses AboutBoxUnit,
{$IFDEF USE_RX_LIBRARY}
  Rxconst,   // For RX_VERSION.
{$ENDIF}
  shellapi,
  shlobj,
  StrUtils,
  MaskUtils;

{$R *.dfm}


type
  TAddWindowTextToListParam = record
    rList: TStrings;
    bWindowMustBeVisible: Boolean;
    hMainForm: HWND;
    hApplication: HWND;
  end;
  TAddWindowHandlesToListParam = record
    strSearchText: String;
    strFoundWindowText: String;
    bCaseSensitive: Boolean;
    bWindowMustBeVisible: Boolean;
    bWholeWindowTextMatchOnly: Boolean;
    rList: TStrings;
    hMainForm: HWND;
    hApplication: HWND;
  end;


// Inspired by HexToInt, posted by Mike Heydon on www.delphi3000.com
function HexToHWND(const HexStr: String): HWND;
var
  i: Integer;
begin
  Result := 0;
  for i := 1 to Length(HexStr) do
  begin
    if Result > High(HWND)/16 then
    begin
      Result := 0;
      break
    end
    else
    begin
      Result := Result * 16;
      if HexStr[i] in ['0'..'9'] then
        Inc(Result, Byte(HexStr[i]) - Byte('0'))
      else
        if HexStr[i] in ['A'..'F'] then
          Inc(Result, 10 + Byte(HexStr[i]) - Byte('A'))
        else
          if HexStr[i] in ['a'..'f'] then
            Inc(Result, 10 + Byte(HexStr[i]) - Byte('a'))
          else
          begin
            // Error: HexStr is not a correct hexadecimal string.
            Result := 0;
            break
          end
    end
  end
end;


function GetWindowsDirWithTrailingBackslash: string;
var
  uLength: UINT;
begin
  try
    SetLength(Result, MAX_PATH + 1);
    SetLastError(0);
    uLength := GetWindowsDirectory(PChar(Result), MAX_PATH);
    if (uLength > 0) and (GetLastError() = 0) then
    begin
      if uLength > MAX_PATH then
      begin
        // Result was not big enough to hold the path.
        SetLength(Result, uLength + 1);
        SetLastError(0);
        uLength := GetWindowsDirectory(PChar(Result), uLength);
        if (uLength > 0) and (GetLastError() = 0) then
          Result := IncludeTrailingBackslash( PChar(Result) )
        else
          // Second call to GetWindowsDirectory failed.
          Result := ''
      end
      else
        Result := IncludeTrailingBackslash( PChar(Result) )
    end
    else
      // First call to GetWindowsDirectory failed.
      Result := ''
  except
    Result := ''
  end;
end;

function GetFirstDroppedFilename(const hd: HDROP): string;
var
  uCountFiles: UINT;
  uFilenameLength: UINT;
begin
  uCountFiles := DragQueryFile(hd, $FFFFFFFF, nil, 0);
  if uCountFiles = 0 then
    Result := ''
  else
  begin
    uFilenameLength := DragQueryFile(hd, 0, nil, 0);
    SetLength(Result, uFilenameLength + 1);
    DragQueryFile(hd, 0, PChar(Result), Length(Result) );
    Result := PChar(Result);
  end;
end;


// Deletes the pair of double-quotes that is enclosing the sting.
function DeleteEnclosingDoubleQuotes(const s: String): String;
begin
  if Length(s) > 1 then
    if (s[1] = '"') and (s[Length(s)] = '"') then
      Result := Copy(s, 2, Length(s) - 2)
    else
      Result := s
  else
    Result := s
end;

// Returns the double-quoted version of a string.
function DoubleQuotedStr(const s: string): string;
begin
  if Pos('"', s) = 0 then
    // The string will be enclosed by double-quotes:
    Result := '"' + s + '"'
  else
    // The string already has some double-quote(s)!
    Result := s
end;


// Puts s at the end of the list.
procedure PutStringAtEndOfList(
  const rList: TStrings;
  const s: string;
  const bOverwrite: Boolean);
var
  iCount: Integer;
begin
  iCount := rList.Count;
  if iCount = 0 then
    rList.Add(s)
  else
    if bOverwrite then
      rList[iCount - 1] := s
    else
      if rList[iCount - 1] = '' then
        rList[iCount - 1] := s
      else
        if rList[iCount - 1] <> s then
          rList.Add(s)
end;


function FindListItemByDisplayName(
  const strName: string;
  const rListView: TShellListView;
  out bSameText: Boolean):
  TListItem;
var
  iIndex: integer;
  rItems: TListItems;
  strFoundName: string;
begin
  rItems := rListView.Items;
  Result := nil;
  bSameText := False;

  for iIndex := 0 to rItems.Count - 1 do
  begin
    strFoundName := Trim(rListView.Folders[iIndex].DisplayName);
    if SameText(strFoundName, strName) then
    begin
      bSameText := True;
      Result := rItems[iIndex];
      Exit
    end
    else
      if Result = nil then
        if AnsiStartsText(strName, strFoundName) then
          Result := rItems[iIndex]
  end
end;


procedure ScrollMemoDown(const rMemo: TMemo);
begin
  SendMessage(rMemo.Handle, WM_VSCROLL,
    MakeLong(SB_THUMBPOSITION, rMemo.Lines.Count - 1), 0 );
end;


function WindowSizeIsZero(const hWindow: HWND): Boolean;
var
  Rect: TRect;
begin
  Rect.Left := 0;
  Rect.Right := 0;
  Rect.Top := 0;
  Rect.Bottom := 0;
  GetWindowRect(hWindow, Rect);

  Result := (Rect.Top = Rect.Bottom) and (Rect.Left = Rect.Right)
end;


// Drops the file on the window by posting a DROPFILES message.
// Returns true iff succesfully posted.  Unfortunately the window might
// still ignore the message, even if it is successfully posted.
// Adapted from DropArgumentsOnWindow, OnTheRun.pas version 1.1.2.2
function DropFileOnWindow(
  const strFile: String;
  const hWindow: HWND): Boolean;
var
  dwBytes: DWORD;
  hGlob: HGLOBAL;
  pDrop: PDropFiles;
  dwError: DWORD;
begin
  Result := False;
  if hWindow <> 0 then
  begin
    // The number of bytes needed for the drop structure, including
    // the zero-terminated filename and a terminating zero.
    dwBytes := SizeOf(TDropFiles) + Length(strFile) + 2;
    hGlob := GlobalAlloc(GMEM_SHARE or GMEM_MOVEABLE or GMEM_ZEROINIT,
      dwBytes);
    if hGlob = 0 then
      Application.MessageBox(
        PChar('Failed to allocate ' + IntToStr(dwBytes) +
          ' bytes from the global heap.'#13#10 +
          SysErrorMessage( GetLastError() ) ),
        'Error preparing to drop file on window',
        MB_ICONEXCLAMATION)
    else
    begin
      pDrop := PDropFiles( GlobalLock(hGlob) );
      if pDrop = nil then
      begin
        dwError := GetLastError();
        GlobalFree(hGlob);
        Application.MessageBox(
          PChar('Failed to lock a global memory object.'#13#10 +
            SysErrorMessage(dwError) ),
          'Error preparing to drop file on window',
          MB_ICONEXCLAMATION)
      end
      else
      begin
        // Copy the filename to the end of the drop structure.
        pDrop.pFiles := SizeOf(TDropFiles);
        StrCopy(PChar(pDrop) + SizeOf(TDropFiles), PChar(strFile));
        GlobalUnlock(hGlob);
        // By now, HDROP(hGlob) is a nice drop handle, ready to be sent!
        // However, SendMessage didn't work, so let's post the drop!
        if PostMessage(hWindow, WM_DROPFILES, WPARAM(hGlob), 0) then
          Result := True
        else
        begin
          dwError := GetLastError();
          GlobalFree(hGlob);
          Application.MessageBox(
            PChar('Failed to post a message.'#13#10 +
              SysErrorMessage(dwError) ),
            'Error trying to drop file on window',
            MB_ICONEXCLAMATION)
        end
      end
    end
  end
end;


// This is an EnumWindowsProc callback function to add the window text
// (or caption) of all open windows to a list of strings.  It will not
// add any empty string or duplicate to the list.
function EnumProcToAddWindowTextToList(
  hwnd: HWND; lParam: LPARAM): BOOL; export; stdcall;
type
  TPParam = ^TAddWindowTextToListParam;
var
  strText: string;
  pParam: TPParam;
begin
  pParam := TPParam(lParam);
  Result := pParam <> nil;

  if Result and (hwnd <> 0) then
    if (hwnd <> pParam.hApplication) and (hwnd <> pParam.hMainForm) then
      if not IsChild(pParam.hMainForm, hwnd) then
      begin
        if pParam.bWindowMustBeVisible then
        begin
          if not IsWindowVisible(hwnd) then
            Exit;
          if WindowSizeIsZero(hwnd) then
            Exit;
        end;
        SetLength(strText, GetWindowTextLength(hwnd) + 1);
        SetLastError(0);
        GetWindowText(hwnd, PChar(strText), Length(strText));
        if GetLastError() = 0 then
        begin
          strText := PChar(strText);
          if strText <> '' then
            if pParam.rList.IndexOf(strText) < 0 then
              pParam.rList.Add(strText)
          end
        end
end;

// This is an EnumWindowsProc callback function to add the window handle
// of all matching open windows to a list of strings.
function EnumProcToAddWindowHandlesToList(
  hwnd: HWND; lParam: LPARAM): BOOL; export; stdcall;
type
  TPParam = ^TAddWindowHandlesToListParam;
var
  strText: string;
  iLength: Integer;
  pParam: TPParam;
  bWindowTextContainsSearchText: Boolean;
begin
  pParam := TPParam(lParam);
  Result := pParam <> nil;
  if Result and (hwnd <> 0) then
    if (hwnd <> pParam.hApplication) and (hwnd <> pParam.hMainForm) then
      if not IsChild(pParam.hMainForm, hwnd) then
      begin
        if pParam.bWindowMustBeVisible then
        begin
          if not IsWindowVisible(hwnd) then
            Exit;
          if WindowSizeIsZero(hwnd) then
            Exit;
        end;

        iLength := GetWindowTextLength(hwnd);
        SetLength(strText, iLength + 1);
        SetLastError(0);
        GetWindowText(hwnd, PChar(strText), Length(strText));
        if GetLastError() = 0 then
        begin
          strText := PChar(strText);

          if pParam.strSearchText = '' then
            bWindowTextContainsSearchText := True
          else
            if pParam.bCaseSensitive then
              bWindowTextContainsSearchText :=
                Pos(pParam.strSearchText, strText) > 0
            else
              bWindowTextContainsSearchText :=
                Pos(UpperCase(pParam.strSearchText), UpperCase(strText)) > 0;

          if bWindowTextContainsSearchText then
            if not pParam.bWholeWindowTextMatchOnly or
              (Length(strText) = Length(pParam.strSearchText)) then
            begin
              pParam.strFoundWindowText := strText;
              pParam.rList.Add(IntToHex(hwnd, 8));
            end
        end
      end
end;


////////////////////////////////////////////////////////////////////////
// Private Methods


// Synchronizes the shell list view with the combo box.
// Returns an empty string if successful, and an error message otherwise.
function TForm1.SyncShellListViewWithFile(const strFile: string): string;
var
  strExpandedFileName: string;
  strExtractedFileName: string;
  strFilePath: string;
  bSameText: Boolean;
  rItem: TListItem;
begin
  if DeleteEnclosingDoubleQuotes(strFile) = '' then
    result := ''
  else
  begin
    strExpandedFileName := ExpandFileName(
      DeleteEnclosingDoubleQuotes(strFile));

    if strExpandedFileName[Length(strExpandedFileName)] = '\' then
      if DirectoryExists(strExpandedFileName) then
      begin
        ShellTreeView1.Path := strExpandedFileName;
        if ShellListView1.Items.Count = 0 then
          ShellListView1.ItemFocused := nil
        else
          ShellListView1.ItemFocused := ShellListView1.Items[0];
        result := ''
      end
      else
        result := 'Directory ' + DoubleQuotedStr(strExpandedFileName) + ' does not exist'
    else
    begin
      strFilePath := ExtractFilePath(strExpandedFileName);
      if DirectoryExists(strFilePath) then
      begin
        ShellTreeView1.Path := strFilePath;
        strExtractedFileName := ExtractFileName(strExpandedFileName);
        if strExtractedFileName <> '' then
        begin
          rItem := FindListItemByDisplayName(
            strExtractedFileName, ShellListView1, bSameText);

          if rItem = nil then
          begin
            ShellListView1.Selected := nil;
            ShellListView1.ItemFocused := nil;
            result := 'Cannot find ' + DoubleQuotedStr(strExtractedFileName)
          end
          else
          begin
            ShellListView1.ItemFocused := rItem;
            rItem.MakeVisible(true);
            if bSameText then
              rItem.Selected := True
            else
              ShellListView1.Selected := nil;

            result := ''
          end
        end
      end
      else
      begin
        result :=
          'Containing directory ' + DoubleQuotedStr(strFilePath) + ' does not exist';
        ShellListView1.ItemFocused := nil;
        ShellListView1.Selected := nil
      end
    end
  end
end;


procedure TForm1.AddListOfMatchingWindowHandlesToComboBox;
var
  Param: TAddWindowHandlesToListParam;
  strStatus: String;
  bToplevelWindowsOnly: Boolean;
  iCount: Integer;
begin
  Param.strSearchText := ComboBoxWindow.Text;
  Param.strFoundWindowText := '';
  ComboBoxHandle.Clear();

  Param.rList := ComboBoxHandle.Items;
  Param.bCaseSensitive := MenuItemCaseSensitiveWindowText.Checked;
  Param.bWindowMustBeVisible := MenuItemDropOnVisibleWindowsOnly.Checked;
  Param.bWholeWindowTextMatchOnly := MenuItemMatchWholeWindowTextOnly.Checked;
  Param.hMainForm := Handle;
  Param.hApplication := Application.Handle;

  bToplevelWindowsOnly := MenuItemDropOnlyOnToplevelWindows.Checked;

  if bToplevelWindowsOnly then
    EnumWindows(@EnumProcToAddWindowHandlesToList, LPARAM(@Param))
  else
    EnumChildWindows(GetDesktopWindow(),
      @EnumProcToAddWindowHandlesToList, LPARAM(@Param));

  iCount := Param.rList.Count;
  ComboBoxHandle.Enabled := iCount <> 0;

  if iCount = 1 then
    strStatus :=
      'Window ' + Param.rList[0] + ': ' + DoubleQuotedStr(Param.strFoundWindowText)
  else
  begin
    strStatus := IntToStr(iCount);
    if Param.bWindowMustBeVisible then
      strStatus := strStatus + ' visible';
    if bToplevelWindowsOnly then
      strStatus := strStatus + ' top-level';

    strStatus := strStatus + ' windows match';

    if Param.bCaseSensitive then
      strStatus := strStatus + ' case sensitively';

    if Param.bWholeWindowTextMatchOnly then
      strStatus := strStatus + ' the whole text '
    else
      strStatus := strStatus + ' the text ';

     strStatus := strStatus + DoubleQuotedStr(Param.strSearchText)
  end;

  PutStringAtEndOfList(Memo1.Lines,strStatus, FOverwriteStatusText);
  FOverwriteStatusText := True;
  ScrollMemoDown(Memo1)
end;


procedure TForm1.Drop(
  const strFile: string;
  const bShowErrorMessageBox: Boolean);
var
  strText: string;
  strStatus: string;
  hWindow: HWND;
  strHandleAsHex: string;
  rList: TStrings;
  iNumWindows: integer;
  bDropped: Boolean;
begin
  bDropped := False;
  PutStringAtEndOfList(Memo1.Lines, '', FOverwriteStatusText);
  ScrollMemoDown(Memo1);
  strHandleAsHex := ComboBoxHandle.Text;
  if strHandleAsHex = '' then
  begin
    AddListOfMatchingWindowHandlesToComboBox();
    rList := ComboBoxHandle.Items;
    iNumWindows := rList.Count;
    if iNumWindows = 1 then
      strHandleAsHex := rList[0]
    else
    begin
      if bShowErrorMessageBox then
        if iNumWindows = 0 then
          Application.MessageBox('There is no target window!',
            'AutoDrop - No target window',
            MB_ICONEXCLAMATION)
        else
          Application.MessageBox(
            PChar('There are ' + IntToStr(iNumWindows) + ' possible target windows!'#13#10 +
            'Please be more specific!'),
            'AutoDrop - More than 1 possible target window',
            MB_ICONEXCLAMATION);

      Exit
    end;
  end;
  hWindow := HexToHWND(strHandleAsHex);

  if hWindow = NULL then
    strStatus := 'Invalid window handle: ' + strHandleAsHex
  else
  begin
    SetLength(strText, GetWindowTextLength(hWindow) + 1);
    SetLastError(0);
    GetWindowText(hWindow, PChar(strText), Length(strText));
    if GetLastError() = 0 then
    begin
      strText := PChar(strText);
      if DropFileOnWindow(strFile, hWindow) then
      begin
        bDropped := True;
        if FileExists(strFile) then
          strStatus := 'File ' + DoubleQuotedStr(ExtractFileName(strFile))
        else
          strStatus := DoubleQuotedStr(strFile);

        strStatus := strStatus + ' dropped on window ' +
          DoubleQuotedStr(strText)
      end
      else
        strStatus := 'Failed to drop arguments on window ' +
          DoubleQuotedStr(strText)
    end
    else
      strStatus := 'Failed to get window text for window ' + strHandleAsHex + '.';

    if not bDropped then
      if bShowErrorMessageBox then
        Application.MessageBox(
          PChar(strStatus),
          'AutoDrop - Error',
          MB_ICONEXCLAMATION);

    PutStringAtEndOfList(Memo1.Lines, strStatus, FOverwriteStatusText);
    FOverwriteStatusText := False;
    ScrollMemoDown(Memo1)
  end;
end;


procedure TForm1.AddListOfWindowsToComboBox;
var
  Param: TAddWindowTextToListParam;
begin
  Param.rList := ComboBoxWindow.Items;
  Param.rList.Clear();
  Param.bWindowMustBeVisible := MenuItemDropOnVisibleWindowsOnly.Checked;
  Param.hMainForm := Handle;
  Param.hApplication := Application.Handle;
  if MenuItemDropOnlyOnToplevelWindows.Checked then
    EnumWindows(@EnumProcToAddWindowTextToList, LPARAM(@Param))
  else
    EnumChildWindows(GetDesktopWindow(),
      @EnumProcToAddWindowTextToList, LPARAM(@Param))
end;


////////////////////////////////////////////////////////////////////////
// Event Handlers

procedure TForm1.ApplicationEvents1Idle(Sender: TObject;
  var Done: Boolean);
begin
  try
    if ParamCount() > 0 then
      EditFile.Text := ParamStr(1);

    AddListOfWindowsToComboBox();
    AddListOfMatchingWindowHandlesToComboBox();

    SyncShellListViewWithFile(EditFile.Text);

    EditFile.OnChange := EditFileChange;
    ShellListView1.OnChange := ShellListView1Change;
    ShellTreeView1.OnChange := ShellTreeView1Change;
    ComboBoxHandle.OnChange := ComboBoxHandleChange;
    ComboBoxWindow.OnChange := ComboBoxWindowChange;

    DragAcceptFiles(Handle, True);
    Cursor := crDefault;
  finally
    Enabled := True;
    Application.OnIdle := nil
  end
end;


procedure TForm1.FormCreate(Sender: TObject);
begin
  // This should have been done in DFM-file!  :-)
  ShellListView1.OnChange := nil;
  ShellTreeView1.OnChange := nil;
  EditFile.OnChange := nil;
  ComboBoxHandle.OnChange := nil;
  ComboBoxWindow.OnChange := nil;

  FOldWidthOfLeftPanel := PanelLeft.Width;
  FDefaultHeightOfStatusBar := PanelBottom.Height;
  Application.OnIdle := ApplicationEvents1Idle
end;


procedure TForm1.MenuItemLargeIconsClick(Sender: TObject);
begin
  ShellListView1.ViewStyle := vsIcon
end;

procedure TForm1.MenuItemSmallIconsClick(Sender: TObject);
begin
  ShellListView1.ViewStyle := vsSmallIcon 
end;

procedure TForm1.MenuItemListClick(Sender: TObject);
begin
  ShellListView1.ViewStyle := vsList
end;

procedure TForm1.MenuItemDetailsClick(Sender: TObject);
begin
  ShellListView1.ViewStyle := vsReport
end;

procedure TForm1.MenuViewClick(Sender: TObject);
begin
  MenuItemStatusBar.Checked :=
    PanelBottom.Height = FDefaultHeightOfStatusBar;
  MenuItemFolders.Checked := PanelLeft.Width > 1;
  MenuItemHiddenFilesAndFolders.Checked :=
    otHidden in ShellListView1.ObjectTypes;

  MenuItemReadOnly.Checked := ShellListView1.ReadOnly;

  MenuItemLargeIcons.Checked := False;
  MenuItemSmallIcons.Checked := False;
  MenuItemList.Checked := False;
  MenuItemDetails.Checked := False;
  case ShellListView1.ViewStyle of
    vsIcon:
      MenuItemLargeIcons.Checked := True;
    vsSmallIcon:
      MenuItemSmallIcons.Checked := True;
    vsList:
      MenuItemList.Checked := True;
    vsReport:
      MenuItemDetails.Checked := True;
  end
end;

procedure TForm1.ShellListView1Change(Sender: TObject; Item: TListItem;
  Change: TItemChange);
var
  rSelectedFolder: TShellFolder;
  strFile: string;
  iAttr: integer;
  strOldText: string;
begin
  if Item <> nil then
  begin
    PutStringAtEndOfList(Memo1.Lines, '', FOverwriteStatusText);
    ScrollMemoDown(Memo1);
    EditFile.OnChange := nil;
    try
      rSelectedFolder := ShellListView1.SelectedFolder();
      if rSelectedFolder <> nil then
      begin
        strFile := rSelectedFolder.PathName();

        if strFile <> '' then
        begin
          strOldText := EditFile.Text;
          if strFile <> strOldText then
          begin
            EditFile.Text := strFile;

            iAttr := FileGetAttr(strFile);
            if (iAttr and faDirectory) = 0 then
              if CheckBoxAutomatically.Checked then
                Drop(strFile, False)
          end
        end
      end
    finally
      EditFile.OnChange := EditFileChange
    end
  end
end;


procedure TForm1.EditFileChange(Sender: TObject);
var
  strStatus: string;
begin
  try
    ShellListView1.OnChange := nil;
    ShellTreeView1.OnChange := nil;
    strStatus := SyncShellListViewWithFile(EditFile.Text);
    { TODO : If a new file is selected, consider automatically dropping the file! }

    PutStringAtEndOfList(Memo1.Lines, strStatus, FOverwriteStatusText);
    FOverwriteStatusText := True;
    ScrollMemoDown(Memo1)
  finally
    ShellTreeView1.OnChange := ShellTreeView1Change;
    ShellListView1.OnChange := ShellListView1Change
  end;
end;

procedure TForm1.MenuItemDropClick(Sender: TObject);
begin
  Drop(EditFile.Text, True)
end;

procedure TForm1.MenuItemAboutClick(Sender: TObject);
begin
  try
    if AboutBox = nil then
      AboutBox := TAboutBox.Create(nil);

{$IFDEF USE_RX_LIBRARY}
    AboutBox.AdditionalVersionInfo :=
      'Used Delphi component library:'#13#10 +
      '  RXLib ' +
      IntToStr(RX_VERSION div $10000) + '.' +
      IntToStr(RX_VERSION mod $10000) +
      ' by Fedor Kozhevnikov, Igor Pavluk, and Serge Korolev'#13#10
{$IFDEF VER140}
      + '  RX Library port to Delphi 6 from http://www.oxygensoftware.com'#13#10
{$ENDIF}
{$IFDEF VER150}
      + '  RX Library port to Delphi 7 from http://www.oxygensoftware.com'#13#10
{$ENDIF}
      + 'Settings are stored at ' +
      DoubleQuotedStr(GetWindowsDirWithTrailingBackslash() + FormStorage1.IniFileName)
      ;
{$ENDIF}
    AboutBox.ShowModal();
  finally
    FreeAndNil(AboutBox)
  end
end;


procedure TForm1.MenuItemFoldersClick(Sender: TObject);
var
  iWidth: Integer;
begin
  iWidth := PanelLeft.Width;
  if iWidth > 1 then
  begin
    FOldWidthOfLeftPanel := iWidth;
    PanelLeft.Width := 1
  end
  else
  begin
    PanelLeft.Width := FOldWidthOfLeftPanel;
  end
end;


procedure TForm1.MenuItemExitClick(Sender: TObject);
begin
  Close()
end;

procedure TForm1.MenuItemDropOnVisibleWindowsOnlyClick(Sender: TObject);
begin
  AddListOfWindowsToComboBox();
  AddListOfMatchingWindowHandlesToComboBox()
end;


procedure TForm1.ComboBoxWindowDropDown(Sender: TObject);
begin
  AddListOfWindowsToComboBox()
end;


procedure TForm1.MenuItemExploreClick(Sender: TObject);
var
  strParam: string;
  iReturnValue: Integer;
  rSelectedFolder: TShellFolder;
begin
  rSelectedFolder := ShellListView1.SelectedFolder;
  if rSelectedFolder = nil then
    strParam := '/n,/e,' + DoubleQuotedStr(ShellTreeView1.Path)
  else
    strParam := '/n,/e,/select,' + DoubleQuotedStr(rSelectedFolder.PathName());

  // If ShellExecute fails, its return value is less than or equal to 32.
  iReturnValue := ShellExecute(0,
    'open', PChar('explorer.exe'), PChar(strParam), nil, SW_SHOWNORMAL);

  if iReturnValue <= 32 then
    Application.MessageBox(
      PChar(
        'Failed to execute Windows Explorer'#13#10 +
        SysErrorMessage(iReturnValue) ),
      'AutoDrop',
      MB_ICONEXCLAMATION)
end;

procedure TForm1.PanelLeftResize(Sender: TObject);
begin
  ShellTreeView1.Enabled := ShellTreeView1.Width > 1
end;

procedure TForm1.MenuItemStatusBarClick(Sender: TObject);
begin
  if PanelBottom.Height = FDefaultHeightOfStatusBar then
    PanelBottom.Height := 1
  else
  begin
    PanelBottom.Height := FDefaultHeightOfStatusBar;
    ScrollMemoDown(Memo1)
  end
end;

procedure TForm1.WMDropFiles(var Msg: TWMDropFiles);
begin
  inherited;
  try
    EditFile.Text := GetFirstDroppedFilename(Msg.Drop);
  finally
    DragFinish(Msg.Drop)
  end
end;

procedure TForm1.ButtonDropClick(Sender: TObject);
begin
  Drop(EditFile.Text, True)
end;

procedure TForm1.MenuItemWebsiteClick(Sender: TObject);
begin
  try
    // Avoid an error message in status bar.
    ComboBoxWindow.OnChange := nil;
    EditFile.OnChange := nil;
    ComboBoxWindow.Text := '';
    EditFile.Text := 'http://www.xs4all.nl/~nd/dekkerware/autodrop/index.htm';
    MenuItemOpen.Click()
  finally
    ComboBoxWindow.OnChange := ComboBoxWindowChange;
    EditFile.OnChange := EditFileChange
  end
end;

procedure TForm1.MenuItemOpenClick(Sender: TObject);
var
  iReturnValue: Integer;
  strFile: string;
  strStatus: string;
begin
  strFile := EditFile.Text;
  // If ShellExecute fails, its return value is less than or equal to 32.
  iReturnValue := ShellExecute(0, 'open', PChar(strFile), nil, nil, SW_SHOWNORMAL);

  if iReturnValue <= 32 then
  begin
    strStatus := 'Failed to open ' + DoubleQuotedStr(strFile);
    Application.MessageBox(
      PChar(
        strStatus + #13#10 +
        SysErrorMessage(iReturnValue) ),
      'AutoDrop',
      MB_ICONEXCLAMATION)
  end
  else
    strStatus := 'Opened ' + DoubleQuotedStr(strFile);

  PutStringAtEndOfList(Memo1.Lines, strStatus, FOverwriteStatusText);
  FOverwriteStatusText := False;
  ScrollMemoDown(Memo1)
end;

procedure TForm1.MenuItemBrowseClick(Sender: TObject);
var
  strFile: String;
begin
  OpenDialog1.Title := 'Browse for file to drop';
  strFile := EditFile.Text;
  if strFile = '' then
    OpenDialog1.InitialDir := ShellTreeView1.Path
  else
    if FileExists(strFile) then
    begin
      strFile := StringReplace(strFile, '/', '\', [rfReplaceAll]);
      OpenDialog1.FileName := strFile
    end
    else
    begin
      OpenDialog1.InitialDir := ShellTreeView1.Path;
      if (Pos('/', strFile) = 0) and (strFile[Length(strFile)] <> '\') then
        OpenDialog1.FileName := strFile
      else
        OpenDialog1.FileName := ''
    end;
  if OpenDialog1.Execute() then
  begin
    strFile := OpenDialog1.FileName;
    EditFile.Text := strFile;
    if CheckBoxAutomatically.Checked then
      Drop(strFile, True)
  end
end;

procedure TForm1.ComboBoxWindowChange(Sender: TObject);
begin
  AddListOfMatchingWindowHandlesToComboBox()
end;

procedure TForm1.MenuItemMatchWholeWindowTextOnlyClick(Sender: TObject);
begin
  AddListOfMatchingWindowHandlesToComboBox()
end;

procedure TForm1.MenuItemCaseSensitiveWindowTextClick(Sender: TObject);
begin
  AddListOfMatchingWindowHandlesToComboBox()
end;

procedure TForm1.MenuItemDropOnlyOnToplevelWindowsClick(Sender: TObject);
begin
  AddListOfWindowsToComboBox();
  AddListOfMatchingWindowHandlesToComboBox()
end;


procedure TForm1.EditFileEnter(Sender: TObject);
begin
  ButtonDrop.Default := True
end;

procedure TForm1.EditFileExit(Sender: TObject);
begin
  ButtonDrop.Default := False
end;


procedure TForm1.ShellTreeView1Change(Sender: TObject; Node: TTreeNode);
begin
  EditFile.OnChange := nil;
  try
    EditFile.Text := IncludeTrailingBackslash(ShellTreeView1.Path)
  finally
    EditFile.OnChange := EditFileChange
  end
end;


procedure TForm1.MenuItemHiddenFilesAndFoldersClick(Sender: TObject);
begin
  if otHidden in ShellListView1.ObjectTypes then
  begin
    ShellTreeView1.ObjectTypes := ShellTreeView1.ObjectTypes - [otHidden];
    ShellListView1.ObjectTypes := ShellListView1.ObjectTypes - [otHidden]
  end
  else
  begin
    ShellTreeView1.ObjectTypes := ShellTreeView1.ObjectTypes + [otHidden];
    ShellListView1.ObjectTypes := ShellListView1.ObjectTypes + [otHidden]
  end
end;


procedure TForm1.ComboBoxHandleChange(Sender: TObject);
var
  strHandle: string;
  strStatus: string;
  strText: string;
  hWindow: HWND;
begin
  strHandle := ComboBoxHandle.Text;
  if strHandle = '' then
    strStatus := ''
  else
  begin
    if ComboBoxHandle.Items.IndexOf(strHandle) = -1 then
      strStatus := 'Window handle '
        + DoubleQuotedStr(strHandle) + ' is not in the drop-down list'
    else
    begin
      hWindow := HexToHWND(strHandle);
      SetLength(strText, GetWindowTextLength(hWindow) + 1);
      SetLastError(0);
      GetWindowText(hWindow, PChar(strText), Length(strText));
      if GetLastError() = 0 then
        strStatus := 'Window ' + strHandle + ': ' + DoubleQuotedStr( Pchar(strText))
      else
        strStatus := 'Window ' + strHandle + ' is not found.'
    end
  end;
  PutStringAtEndOfList(Memo1.Lines, strStatus, FOverwriteStatusText);
  FOverwriteStatusText := True;
  ScrollMemoDown(Memo1)
end;


procedure TForm1.MenuItemRefreshClick(Sender: TObject);
begin
  PutStringAtEndOfList(Memo1.Lines, '', FOverwriteStatusText);
  FOverwriteStatusText := True;
  ScrollMemoDown(Memo1);
  try
    Enabled := False;
    ShellListView1.OnChange := nil;
    ShellListView1.Refresh();
    SyncShellListViewWithFile(EditFile.Text);
    AddListOfWindowsToComboBox();
    AddListOfMatchingWindowHandlesToComboBox();
  finally
    ShellListView1.OnChange := ShellListView1Change;
    Enabled := True
  end

end;

procedure TForm1.ComboBoxHandleExit(Sender: TObject);
var
  strHandle: String;
  strStatus: String;
begin
  strHandle := ComboBoxHandle.Text;
  if strHandle <> '' then
    if ComboBoxHandle.Items.IndexOf(strHandle) = -1 then
    begin
      strStatus := 'Window handle '
        + DoubleQuotedStr(strHandle) + ' is not in the drop-down list';

      Application.MessageBox( PChar(
        'Window handle ' + DoubleQuotedStr(strHandle) + ' is not in the drop-down list!'#13#10#13#10 +
        'Please either select a window handle from the list or empty the field.' ),
        'AutoDrop - Window handle not in the drop-down list',
        MB_ICONEXCLAMATION);

      ComboBoxHandle.SetFocus()
    end

end;

procedure TForm1.MenuItemReadOnlyClick(Sender: TObject);
var
  strText: String;
  strStatus: String;
begin
  Enabled := False;
  strText := EditFile.Text;
  ShellListView1.OnChange := nil;
  ShellTreeView1.OnChange := nil;
  try
    ShellListView1.ReadOnly := not ShellListView1.ReadOnly;
    strStatus := SyncShellListViewWithFile(EditFile.Text);
    PutStringAtEndOfList(Memo1.Lines, strStatus, FOverwriteStatusText);
    FOverwriteStatusText := True;
    ScrollMemoDown(Memo1)
  finally
    ShellListView1.OnChange := ShellListView1Change;
    ShellTreeView1.OnChange := ShellTreeView1Change;
    Enabled := True;
  end
end;

end.
